//****************************************************************************************
// Driver program for RCCE functional emulator. 
//****************************************************************************************
// This file will only get linked if the OpenMP-based emulator mode is 
// used, so we know OpenMP is always available. 
//
// Author: Rob F. Van der Wijngaart
//         Intel Corporation
// Date:   12/22/2010
//
//****************************************************************************************
// 
// Copyright 2010 Intel Corporation
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
// 
#include <stdio.h>
#include <stdlib.h>
#include "RCCE_lib.h"
#ifdef RC_POWER_MANAGEMENT
#include "RCCE_lib_pwr.h"
#endif
#include <sys/shm.h>
#include <unistd.h>
#include <string.h>
//#include <iostream>
#include <errno.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/stat.h>        /* For mode constants */
#include <fcntl.h>           /* For O_* constants */

#if 0 
#define debug(var) printf("[%s:%s:%d] %s = \"%s\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#define debug_addr(var) printf("[%s:%s:%d] %s = \"%X\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#define debug_int(var) printf("[%s:%s:%d] %s = \"%d\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#define debug_size_t(var) printf("[%s:%s:%d] %s = \"%zu\"\n", __FILE__, __FUNCTION__, __LINE__, #var, var); fflush(NULL)
#else
#define debug(var)
#define debug_addr(var)
#define debug_int(var)
#define debug_size_t(var)
#endif

//----------------------------------------------------------------------------------------
// global variables used for OpenMP emulation:
// - shared array of openmp locks (emulating single test-and-set register per core)
// - shared MPB space                                         
//----------------------------------------------------------------------------------------
// TODO: Make real replacement
//omp_lock_t  RCCE_corelock[RCCE_MAXNP];
//t_vchar     RC_comm_buffer[RCCE_MAXNP*RCCE_BUFF_SIZE_MAX];
//t_vchar     RC_shm_buffer[RCCE_SHM_SIZE_MAX];
//t_vchar     RCCE_corelock[RCCE_MAXNP];

t_vchar     *RC_comm_buffer;
t_vchar     *RC_shm_buffer;
t_vchar     *RCCE_corelock;
volatile unsigned char *fork_barrier_data;
volatile unsigned char *fork_startup_barrier_data;
unsigned char fork_id;

#define assert_int_eq(got, expected) if(got != expected) { printf("[%s:%s:%d][ERROR] Got %s = %d, expected %s = %d\n", __FILE__, __FUNCTION__, __LINE__, #got, got, #expected, expected); abort(); }
#define assert_int_neq(got, ref) if(got == ref) { printf("[%s:%s:%d][ERROR] Got %s = %d, expected %s = %d\n", __FILE__, __FUNCTION__, __LINE__, #got, got, #ref, ref); perror(NULL); abort(); }
#define assert_size_t_eq(got, expected) if(got != expected) { printf("[%s:%s:%d][ERROR] Got %s = %zu, expected %s = %zu\n", __FILE__, __FUNCTION__, __LINE__, #got, got, #expected, expected); abort(); }
#define assert_size_t_neq(got, ref) if(got == ref) { printf("[%s:%s:%d][ERROR] Got %s = %zu, expected %s = %zu\n", __FILE__, __FUNCTION__, __LINE__, #got, got, #ref, ref); perror(NULL); abort(); }

// name of user application. In non-OpenMP emulation mode this resolves to main. In 
// OpenMP emulation mode it is unchanged
int RCCE_APP(int argc, char **argv);

#ifndef COPPERRIDGE
int RCCE_NPE;
inline int
fork_check_startup_barrier(int ref)
{
	int i, total;
	total = 1;
	for(i = 0; i <= RCCE_NPE; i++)
	{
		total &= fork_startup_barrier_data[i] == ref;
	}

	return total;
}

inline int
fork_startup_barrier(int index)
{
	fork_startup_barrier_data[index] = !fork_startup_barrier_data[index];
	while(!fork_check_startup_barrier(fork_startup_barrier_data[index]));
	return 0;
}
#endif

int main(int argc, char *argv[]){
  int i;
  pid_t ret;
  void* buffer;
#ifdef COPPERRIDGE
int RCCE_NPE;
#endif

  int RCCE_error = 0;
/*
  if (argc <3) {
    printf("RCCE DRIVER ERROR: no executable or parameters specified\n");
    exit(1);
  }
*/

  char *rcce_nues_str = getenv("RCCE_NUES");
  if(rcce_nues_str == NULL || strcmp(rcce_nues_str, "") == 0)
  {
    printf("Missing number of UEs. Please define environment variable RCCE_NUES and try again (example: RCCE_NUES=4 %s ...)\n", argv[0]);
    exit(1);
  }
  RCCE_NP = RCCE_NPE = atoi(rcce_nues_str);
  //RCCE_NP = RCCE_NPE = atoi(*(argv+1));
  if (RCCE_NPE<1 || RCCE_NPE>RCCE_MAXNP) {
    printf("Invalid number of UEs: %d\n", RCCE_NPE);
    exit(1);
  }

//  extern omp_lock_t RCCE_corelock[RCCE_MAXNP];
//  extern t_vchar    RC_comm_buffer[RCCE_MAXNP*RCCE_BUFF_SIZE_MAX];
//  extern t_vchar    RC_shm_buffer[RCCE_SHM_SIZE_MAX];
//  extern unsigned char *fork_barrier_data[RCCE_MAXNP];

  // Size of all buffers
  size_t lock_size = sizeof(t_vchar) * RCCE_MAXNP;
  debug_int(RCCE_MAXNP * RCCE_BUFF_SIZE_MAX);
#ifdef RC_POWER_MANAGEMENT
  size_t comm_buffer_size = sizeof(t_vchar) * RCCE_MAXNP * (RCCE_BUFF_SIZE_MAX + REGULATOR_LENGTH);
#else
  size_t comm_buffer_size = sizeof(t_vchar) * RCCE_MAXNP * RCCE_BUFF_SIZE_MAX;
#endif
  size_t shm_buffer_size = sizeof(t_vchar) * RCCE_SHM_SIZE_MAX;
  size_t fork_barrier_data_size = sizeof(volatile unsigned char) * RCCE_MAXNP;
  size_t fork_startup_barrier_data_size = sizeof(volatile unsigned char) * RCCE_MAXNP + 1;
  size_t total_size = lock_size + comm_buffer_size + shm_buffer_size + fork_barrier_data_size + fork_startup_barrier_data_size;
//  size_t total_size = lock_size + comm_buffer_size + shm_buffer_size + fork_barrier_data_size;

  // Allocate and map shared memory buffers
//  int page_id;
//  page_id = shmget(1, total_size, IPC_CREAT | 0777);
//  assert_neq(page_id, -1);
//  buffer = shmat(page_id, NULL, 0);
//  assert_neq((size_t)buffer, -1);

  int fd;
  fd = shm_open("/mpb", O_CREAT | O_RDWR, S_IRWXU | S_IRWXG | S_IRWXO);
  if(fd < 0)
  {
    printf("[%s:%s:%d] Error while linking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
    fd = shm_unlink("/mpb");
    if(fd < 0)
    {
      printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
    }
    exit(fd);
  }
  assert_int_neq(fd, -1);
  if(ftruncate(fd, total_size) != 0)
  {
    printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
    abort();
  }
  buffer = mmap(NULL, total_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
  assert_size_t_neq((size_t)buffer, (size_t)MAP_FAILED);

  // Give each shared buffer an address
  RCCE_corelock = (t_vchar*)buffer;
  RC_comm_buffer = (t_vchar*)((size_t)buffer + lock_size);
  RC_shm_buffer = (t_vchar*)((size_t)buffer + lock_size + comm_buffer_size);
  fork_barrier_data = (volatile unsigned char*)((size_t)buffer + lock_size + comm_buffer_size + shm_buffer_size);
  fork_startup_barrier_data = (volatile unsigned char*)((size_t)buffer + lock_size + comm_buffer_size + shm_buffer_size + fork_barrier_data_size);

  // Fork all id
  for(i = 0; i < RCCE_MAXNP; i++)
  {
    fork_barrier_data[i] = 0;
    fork_startup_barrier_data[i] = 0;
  }
  // have one more element to initialize in startup barrier
  fork_startup_barrier_data[i] = 0;

  // initialize comm buffers with all constants 
  for (i=0; i<RCCE_MAXNP*RCCE_BUFF_SIZE_MAX/sizeof(int); i++) 
    *(((int *) RC_comm_buffer)+i) = RCCE_comm_init_val;

  // initialize core locks 
  for (i=0; i<RCCE_NPE; i++) RCCE_corelock[i] = 0;

  // If we need to debug, prepare to printf all child pid and pause the program
  char* debug_str = NULL;
  int debug = 0;
  debug_str = getenv("DEBUG");
  if(debug_str != NULL)
  {
    if(strcmp(debug_str, "0") != 0 && strcmp(debug_str, "") != 0)
    {
      debug = 1;
    }
    else
    {
          printf("Debug is not activated\n");
    }
  }
  else
  {
    printf("Environment variable \"DEBUG\" was not set. Debug is not activated\n");
  }

	int *pids = malloc(sizeof(int) * RCCE_NPE);
	
  // start parallel region. If any UE (thread) returns a non-zero error
  // value, the total error, reduced to RCCE_error, will be non-zero
  if(RCCE_NPE > 1)
  {
    for(i = 0; i < RCCE_NPE; i++)
    {
      fork_id = i;
      ret = fork();

#if 0
	// Duplicate the argument stack
	char **local_argv = malloc(sizeof(char*) * (argc + 1));
	size_t j;
	for(j = 0; j < argc; j++)
	{
		local_argv[j] = malloc(sizeof(char) * (strlen(argv[j]) + 1));
		strcpy(local_argv[j], argv[j]);
		local_argv[j][strlen(argv[j])] = '\0';
	}
	local_argv[argc] = NULL;
#endif

      if(ret == 0)
      {
        // get shared memory buffers
        fd = shm_open("/mpb", O_RDWR, S_IRWXU | S_IRWXG | S_IRWXO);
        if(fd < 0)
        {
          printf("[%s:%s:%d] Error while linking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
          fd = shm_unlink("/mpb");
          if(fd < 0)
          {
            printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
          }
          exit(fd);
        }
        assert_int_neq(fd, -1);
        if(ftruncate(fd, total_size))
	{
          printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
    	  abort();
	}
        buffer = mmap(NULL, total_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
        assert_size_t_neq((size_t)buffer, (size_t)MAP_FAILED);

        // Give each shared buffer an address
        RCCE_corelock = (t_vchar*)buffer;
        RC_comm_buffer = (t_vchar*)((size_t)buffer + lock_size);
        RC_shm_buffer = (t_vchar*)((size_t)buffer + lock_size + comm_buffer_size);
        fork_barrier_data = (t_vchar*)((size_t)buffer + lock_size + comm_buffer_size + shm_buffer_size);
	fork_startup_barrier_data = (t_vchar*)((size_t)buffer + lock_size + comm_buffer_size + shm_buffer_size + fork_barrier_data_size);

	// Wait for all cores and parent process to be ready
	fork_startup_barrier(i);


#if 0
	// Start virtual UEs (processes)
        RCCE_error = RCCE_APP(argc, local_argv);
	// Destroy the argument stack
	for(j = 0; j < argc; j++)
	{
		free(local_argv[j]);
	}
	free(local_argv);
#else
	// Start virtual UEs (processes)
        RCCE_error = RCCE_APP(argc, argv);
#endif

        return RCCE_error;
      }
      else
      {
        if(ret < 0)
        {
          return RCCE_ERROR_NUM_UES;
        }
        else
        {
          if(debug)
          {
            printf("Core %d has pid %d\n", i, ret);
          }

	  pids[i] = ret;
        }
      }
    }

    // Let the cores begin
   char pause[1];
   if(debug)
   {
     printf("Hit enter to start cores\n");
	int res;
	while((res = fgetc(stdin)) != 10 && res != EOF);
     if(res != 1 || res == EOF)
     {
       printf("[%s:%s:%d] Error while waiting for user to unlock processes (type enter only). Aborting\n", __FILE__, __FUNCTION__, __LINE__);
       abort();
     }
   }
   // Let all cores start
   fork_startup_barrier(RCCE_NPE);

   if(debug)
   {
     printf("%d cores started\n", RCCE_NPE);
   }

  int exit_val = 0;
  int status;
  for(i = 0; i < RCCE_NPE; i++)
  {
    int pid = wait(&status);

    int j;
    for(j = 0; j < RCCE_NPE; j++)
    {
	if(pids[j] == pid)
	{
		printf("Core %i (pid %i) exits with value %i\n", j, pid, status);
	}
    }
    exit_val = exit_val | status;
  }

  fd = shm_unlink("/mpb");
  if(fd != 0)
  {
    printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator. Aborting.\n", __FILE__, __FUNCTION__, __LINE__);
    exit(fd);
  }
  exit(exit_val);
  }
  else
  {
   if(debug)
   {
   char pause[1];
            printf("Core 0 has pid %d\n", getpid());
     printf("Hit enter to start cores\n");
	int res;
	while((res = fgetc(stdin)) != 10 && res != EOF);
     if(res == EOF)
     {
       printf("[%s:%s:%d] Error while waiting for user to unlock processes (type enter only). Aborting\n", __FILE__, __FUNCTION__, __LINE__);
       abort();
     }
   }
        // get shared memory buffers
        fd = shm_open("/mpb", O_RDWR, S_IRWXU | S_IRWXG | S_IRWXO);
        if(fd < 0)
        {
          printf("[%s:%s:%d] Error while linking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
          if(fd < 0)
          {
            printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
          }
          exit(fd);
        }
        assert_int_neq(fd, -1);
        if(ftruncate(fd, total_size))
	{
          printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator with message \"%s\". Aborting.\n", __FILE__, __FUNCTION__, __LINE__, strerror(errno));
	  abort();
	}
        buffer = mmap(NULL, total_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
        assert_size_t_neq((size_t)buffer, (size_t)MAP_FAILED);

        // Give each shared buffer an address
        RCCE_corelock = (t_vchar*)buffer;
        RC_comm_buffer = (t_vchar*)((size_t)buffer + lock_size);
        RC_shm_buffer = (t_vchar*)((size_t)buffer + lock_size + comm_buffer_size);
        fork_barrier_data = (t_vchar*)((size_t)buffer + lock_size + comm_buffer_size + shm_buffer_size);
	fork_startup_barrier_data = (t_vchar*)((size_t)buffer + lock_size + comm_buffer_size + shm_buffer_size + fork_barrier_data_size);

	// Start virtual UEs (processes)
        RCCE_error = RCCE_APP(argc, argv);
	printf("Core 0 exits with value %i\n", RCCE_error);
  	fd = shm_unlink("/mpb");
        if(fd != 0)
        {
          printf("[%s:%s:%d] Error while unlinking shared memory MPB emulator. Aborting.\n", __FILE__, __FUNCTION__, __LINE__);
          exit(fd);
        }
   	exit(RCCE_error);
   }
}

// Add empty SCC API function shells
#ifdef SHMADD_CACHEABLE
int DCMflush() {
   // Do nothing
   return 1;
}
#endif

